%%%
%%% Shell
%%%

functor
   
import
   Module Open OS Pickle Property Resolve

   XApplication(wrapper:AppMan parseCmdArgs:ParseCmdArgs)
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   XTime(localTimeStamp)
   at 'x-ozlib://mathweb/share/XTime.ozf'
   Env(home:Home)
   at 'x-ozlib://mathweb/share/Env.ozf'
   File(exists loadVS)
   at 'x-ozlib://mathweb/share/File.ozf'
   Text(forAll make)
   at 'x-ozlib://mathweb/share/Text.ozf'
   ShellCommands(addAll)
   at 'x-ozlib://mathweb/mosh/ShellCommands.ozf'
   TextServer('class':Server)
   at 'x-ozlib://mathweb/mosh/TextServer.ozf'
   
export
   generic: Shell
   'local': LocalShell
   remote:  RemoteShell
   
define
   
   local
      
      class TextFile from Open.file Open.text end

      V2VS = Value.toVirtualString
      
      VS2A = VirtualString.toAtom
      VS2S = VirtualString.toString
 
      fun {Enumerate Xs Sep}
         fun {DoEnumerate Xs}
            case Xs of nil then ""
            [] X|Xr then
               Sep#X#{DoEnumerate Xr}
            end
         end
      in
         case Xs of nil then ""
         [] X|Xr then
            X#{DoEnumerate Xr}
         end
      end
      
      %% search for existing file
      fun {FindFile FileNames}
         case FileNames
         of nil  then false
         [] F|Fr then
            try
               {OS.stat F stat(type:reg...)} F
            catch _ then
               {FindFile Fr}
            end
         end
      end
      
      %%
      %% analyze command line
      %%

      fun {ReadSpaces S} 
         case S of nil then nil
         [] C|Sr then
            if {Char.isSpace C}
            then {ReadSpaces Sr}
            else S
            end
         end
      end
      proc {ReadTillSpace S ?Str ?Rest}
         case S of nil then Str=nil Rest=nil
         [] C|Sr then
            if {Char.isSpace C}
            then Str=nil Rest=S 
            else Str=C|{ReadTillSpace Sr $ ?Rest}
            end
         end
      end
      proc {ReadTillEndOfVarName S ?Str ?Rest}
         case S of nil then Str=nil Rest=nil
         [] C|Sr then
            if {Char.isSpace C} orelse {Member C [&. &# &| &) &} ]}
            then Str=nil Rest=S 
            else Str=C|{ReadTillEndOfVarName Sr $ ?Rest}
            end
         end 
      end
      
      %% separate command & args    
      proc {GetCommand S ?CommandS ?RestS}
         {ReadTillSpace {ReadSpaces S} ?CommandS ?RestS}
      end
      fun {GetArgs S}
         S1 = {ReadSpaces S}
      in
         if S1 == nil then nil
         else
            FirstS RestS in
            {ReadTillSpace S1 ?FirstS ?RestS}
            FirstS|{GetArgs RestS}
         end
      end
      
      %% comments
      fun {IsCommentLine S}
         case {ReadSpaces S} of
            nil  then true
         [] &#|_ then true
         [] &%|_ then true
         else false
         end
      end
      
      %% backquotes
      local
         proc {ReadTillBackquote S ?Str ?Rest ?Found}
            case S of nil then Str=nil Rest=nil Found=false
            [] C|Sr then
               if C==&`
               then Str=nil Rest=Sr Found=true
               else Str=C|{ReadTillBackquote Sr $ ?Rest ?Found}
               end
            end   
         end
      in
         proc {SplitBackquotes S ?Prefix ?FirstQuoted ?Suffix ?Found}
            Rest in
            {ReadTillBackquote S ?Prefix ?Rest ?Found}
            if Found == false then
               FirstQuoted=nil Suffix=nil
            elseif {ReadTillBackquote Rest ?FirstQuoted ?Suffix}==false then
               raise parseError('Unmatched `.') end  
            end
         end         
      end

      %% variables $XYZ bzw. ${XYZ}
      local
         proc {ReadTillDollar S ?Str ?Rest}
            case S of nil then Str=nil Rest=false
            [] C|Sr then
               if C==&$ andthen case Sr of nil then false
                                [] & |_ then false
                                [] &)|_ then false
                                [] &}|_ then false
                                else true
                                end
               then Str=nil Rest=Sr
               else Str=C|{ReadTillDollar Sr $ ?Rest}
               end
            end
         end
         proc {ReadTillRightBrace S ?Str ?Rest}
            case S of nil then
               raise parseError('Missing }.') end
            [] C|Sr then
               if C==&}
               then Str=nil Rest=Sr
               else Str=C|{ReadTillRightBrace Sr $ ?Rest}
               end
            end
         end
         proc {ReadVarName S ?VarName ?RestS}
            case S of nil then
               VarName=nil RestS=nil
            [] &{|Sr then
               {ReadTillRightBrace Sr ?VarName ?RestS} 
            else
               RestS = & |{ReadTillEndOfVarName S ?VarName $}
            end
         end
         fun {FindVarRefs S}
            case S of nil then nil
            else
               S1 R1 in
               {ReadTillDollar S ?S1 ?R1}
               case R1 of false then [S1]
               [] &$|&$|R2 then
                  VarName Rest in
                  {ReadVarName R2 ?VarName ?Rest}
                  S1|'$$$'(VarName)|{FindVarRefs Rest}
               [] &$|R2 then
                  VarName Rest in
                  {ReadVarName R2 ?VarName ?Rest}
                  S1|'$$'(VarName)|{FindVarRefs Rest}
               else
                  VarName Rest in
                  {ReadVarName R1 ?VarName ?Rest}
                  S1|'$'(VarName)|{FindVarRefs Rest}
               end
            end
         end      
      in
         fun {Concat Xs}
            case Xs of nil then ""
            [] X|Xr then X#{Concat Xr} end
         end   
         fun {DerefVars S GetEnvFun}
            L = {Map {FindVarRefs S}
                 fun {$ X}
                    case X of '$'(VarName) then
                       case {GetEnvFun VarName} of false then
                          raise undefVar(VarName) end
                       elseof Val then
                          if {IsVirtualString Val}
                          then Val
                          else {V2VS Val 5 20}
                          end
                       end
                    [] '$$'(VarName) then
                       " local "#
                       "    App_ = {App getLocalEnv('"#VarName#"' $)} "#
                       " in "#
                       "    {App_ getLocalEnv('self' $)} "#
                       " end "
                    [] '$$$'(VarName) then
                       " local "#
                       "    APP  = {App getLocalEnv('APP' $)} "#
                       "    App_ = {App getLocalEnv(APP $)} "#
                       " in "#
                       "    {App_ getLocalEnv('"#VarName#"' $)} "#
                       " end "
                    else X
                    end
                 end}
         in
            {VirtualString.toString {Concat L}}
         end
      end
      
      %% output redirection
      local
         proc {TryAssign Var Val}
            if {IsDet Var}
            then raise parseError('Ambiguous redirect.') end
            else Var = Val
            end
         end
         proc {DoCase X Xr Var Label ?XRest}
            if X == nil then
               File|!XRest = Xr in
               {TryAssign Var Label(File)}
            else
               XRest = Xr
               {TryAssign Var Label(X)}
            end
         end
      in
         proc {GetRedirections AllArgs ?Args OutFile ErrFile}
            case AllArgs of nil then
               Args = nil
               if {IsFree OutFile} then OutFile = unit end
               if {IsFree ErrFile} then ErrFile = unit end
            [] X|Xr then
               case X of &>|X1 then
                  XRest in
                  case X1 of &>|&&|X2 then
                     {DoCase X2 Xr ErrFile append ?XRest}
                  elseof &>|X2 then
                     {DoCase X2 Xr OutFile append ?XRest}
                  elseof &&|X2 then
                     {DoCase X2 Xr ErrFile truncate ?XRest}
                  else
                     {DoCase X1 Xr OutFile truncate ?XRest}
                  end
                  {GetRedirections XRest ?Args OutFile ErrFile}
               else
                  Args = X|{GetRedirections Xr $ OutFile ErrFile}
               end
            end
         end
      end

      %% provide GetCmdArgs function
      fun {MakeGetCmdArgs Args}
         fun {$ Spec} {ParseCmdArgs Spec Args} end
      end

      proc {DoCmd CmdName App Proc}
	 try
            {Proc}
         catch error(ap(usage S)...) then
            {App.showError CmdName#': '#S}
         [] system(os(A F I S)...) then
            {App.showError A#' error: '#F#': ['#I#'] '#S}
         end               
      end
   in
      class Shell from AppMan
         
         feat
            name           % application name, e.g. 'mosh'   
            pid            % process ID
            starttimeVS    % time stamp at startup
            directArgs     % direct arguments available as $1, $2, $3, ...
            commands       % dictionary with items: proc {$ AppMan}
            aliases        % dictionary with items: VS
            exitException  % Name
            
         attr
            verbose:      true
            userMessages: nil
            
         meth init(name:       AppName
                   appMan:     App
                   directArgs: DirectArgs <= nil
                   prompt:     Prompt     <= "> "
                   commands:   Commands   <= {NewDictionary}
                   rcFiles:    RCFiles    <= [Home#"/."#AppName#"rc"]
                   verbose:    Verbose    <= true
                   exitExc:    ExitExc    <= {NewName})
            
            AppMan, init(App)
            self.name          = AppName
            self.pid           = {OS.getPID}
            self.starttimeVS   = {XTime.localTimeStamp}  
            self.directArgs    = DirectArgs
            self.commands      = Commands
            self.aliases       = {NewDictionary}
            self.exitException = ExitExc
            
            verbose      <- Verbose
            userMessages <- nil

	    {self addBuiltin('version' BuiltinVersion)}
            {self addBuiltin('alias'   BuiltinAlias)}
            {self addBuiltin('unalias' BuiltinUnalias)}
            {self addBuiltin('source'  BuiltinSource)}
            {self addBuiltin('apply'   BuiltinApply)}
            {self addBuiltin('start'   BuiltinStart)}
            {self addBuiltin('help'    BuiltinHelp)}
            {ShellCommands.addAll Commands}
            
            {Dictionary.put self.aliases '?'      'help'#"--commands"}
            {Dictionary.put self.aliases 'set'    'setenv'#"-l "}
            {Dictionary.put self.aliases 'relink' 'apply'#"relink "}

            UserHome = {Property.get 'user.home'}
            OzHome   = {Property.get 'oz.home'}
            
            if AppMan,getLocalEnv('RESOLVE' $) == false then
               {OS.putEnv 'RESOLVE'
                'cache='#UserHome#'/.oz/cache'#':'#
                'cache='#OzHome#'/cache'}
            end

	    VersionVS = "This is mozart shell version 1.0.0 beta."
	 in
            AppMan,putLocalEnv('_MOSH_VERSION_VS' VersionVS)
            AppMan,putLocalEnv('PROMPT'           fun {$} Prompt end)
            AppMan,putLocalEnv('PID'              self.pid)
            AppMan,putLocalEnv('?'                unit)
            AppMan,putLocalEnv('self'             self)
            AppMan,putLocalEnv('MOSH'             AppMan,object($))
            AppMan,putLocalEnv('APP'              'MOSH')
            
            if @verbose then
               {self showVersion}
            end
            
            if {Dictionary.keys self.commands} == nil then
               AppMan,showInfo("WARNING: No commands specified.")
            end
            
            if AppMan,getLocalEnv('TIMEOUT' $) == false then
               AppMan,putLocalEnv('TIMEOUT' 10)
            end
            
            AppMan,putLocalEnv('*' {VS2S {Enumerate DirectArgs " "}})
            {List.forAllInd DirectArgs
             proc {$ N Val} AppMan,putLocalEnv(N Val) end}
            
	    {ForAll RCFiles
             proc {$ RCFile}
		ExitException = self.exitException in
                try 
		   case {FindFile [RCFile]} of false then
		      {self showInfo("File `"#RCFile#"' not found.")}
		   elseof ScriptFile then
		      if @verbose then
			 {self showInfo("Processing file `"#ScriptFile#"'...")}
		      end
		      {self doFile(ScriptFile)}
		   end
		catch !ExitException then skip end
	     end}
	    
	 end

         %% %%%%%%%%%%%%%%%%%%%%%%% %%
         %% Command Line Processing %%
         %% %%%%%%%%%%%%%%%%%%%%%%% %%

         meth doFile(FileName)
            F  = {New TextFile init(name:FileName)}
            VS = {F read(list:$ size:all)}
         in
            {F close}
            {self doVS(VS)}
         end
         meth doVS(VS)
            {self doText({Text.make VS})}
         end        
         meth doText(Txt)
            {Text.forAll Txt proc {$ Line} {self doLine(Line)} end}
	 end
	 meth doLineSimple(Line)
            CmdS CmdA1 CmdA ArgsS1 ArgsS
            GetEnvFun = fun {$ Key} AppMan,getLocalEnv(Key $) end
         in
            {GetCommand {DerefVars Line GetEnvFun} ?CmdS ?ArgsS1}
            CmdA1 = {StringToAtom CmdS}
            case {Dictionary.condGet self.aliases CmdA1 false} of false then
               CmdA=CmdA1 ArgsS=ArgsS1
            elseof CmdA2#RestS then
               CmdA=CmdA2 ArgsS={Append RestS ArgsS1}
            end
	    if CmdA == 'exit' then
	       GetCmdArgs = {MakeGetCmdArgs {GetArgs ArgsS}}
	       proc {P} {self builtinExit(AppMan,object($) GetCmdArgs)} end
            in
	       {DoCmd 'exit' AppMan,object($) P}
            elseif CmdA == '' then
               skip
            else
               %%case {Dictionary.condGet self.commands CmdA false}
               %%of false then
               %%   raise unknownCommand(CmdA) end
               %%elseof P then
               local
                  Args OutFile ErrFile
                  {GetRedirections {GetArgs ArgsS} ?Args ?OutFile ?ErrFile}
		  Proc0 = case {Dictionary.condGet self.commands CmdA false}
                          of false then
                             proc {$}
                                {self Apply(AppMan,object($) CmdA Args)}
			     end
                          elseof P then
			     GetCmdArgs = {MakeGetCmdArgs Args} in
			     proc {$} {P AppMan,object($) GetCmdArgs} end
			  end
		  {self showDebug("executing `"#CmdA#"' "#
				  "with args `"#ArgsS#"'")}
	       in
		  AppMan,redirectOutputToFile(Proc0
					      stdout:OutFile
					      stderr:ErrFile)
               end
	    end
         end
         meth replaceBackquotes(Line $)
            Prefix FirstQuoted Suffix in
            if {SplitBackquotes Line ?Prefix ?FirstQuoted ?Suffix} == true
            then
               Tmp = {OS.tmpnam}
               proc {ExecQuoted} {self doLineSimple(FirstQuoted)} end
               AppMan,redirectOutputToFile(ExecQuoted
                                           stdout: truncate(Tmp))
               Output = {File.loadVS Tmp}
               {OS.unlink Tmp}
            in
               Prefix#Output#{self replaceBackquotes(Suffix $)}
            else Line
            end
         end
         meth doLine(Line)
            try
               if {IsCommentLine Line}==false then
                  Line1 = {VS2S {self replaceBackquotes(Line $)}} in
                  {self doLineSimple(Line1)}
               end
            catch Err then
               ExitException = self.exitException
            in
               case Err
               of system(kernel(terminate...)...) then
                  raise Err end
               [] !ExitException then
                  raise Err end
               [] disconnect then
                  raise Err end
               [] unknownCommand(CmdA) then
                  {self showError('mosh: '#CmdA#': unknown command')}
               [] parseError(ErrMsg) then
                  {self showError(ErrMsg)}
               [] undefVar(VarName) then
                  {self showError(VarName#': '#'Undefined variable.')}
               else
                  {self printExc(Err)}
               end
            end
         end

         meth showUser(Msg)
            userMessages <- Msg|@userMessages
         end
         
         meth prompt
            PromptFun in
            {Thread.preempt {Thread.this}}
            if @userMessages \= nil then
               {self showInfo('')}
               {ForAll {Reverse @userMessages}
                proc {$ Msg} {self showInfo(Msg)} end}
               userMessages <- nil
            end
            AppMan, getLocalEnv('PROMPT' ?PromptFun)
            {self printInfo({PromptFun})} 
         end
            
         meth loop(InputTextObj)
            proc {DoLoop}
               if {InputTextObj atEnd($)}==false then
                  local Line in
                     {self prompt}
                     {InputTextObj getS(?Line)}
                     if Line\=false then
                        {self doLine(Line)}
                     end
                  end
                  {DoLoop}
               end
            end
         in
            {DoLoop}
         end
         
         %% %%%%%%%%%%%%%%%% %%
         %% Builtin Commands %%
         %% %%%%%%%%%%%%%%%% %%

         %% exit
         meth exit
            {self showInfo("exit")}
         end
	 meth builtinExit(App GetCmdArgs)
            Spec = record(force(single char:&f type:bool default:false))
            Args = {GetCmdArgs Spec}
         in
            case Args.1 of nil then
               if AppMan,isEmpty($) orelse Args.force==true
               then raise self.exitException end
               else {App.showError "There are non-terminated threads."}
               end
            else
               raise error(ap(usage 'unexpected arguments')) end
            end
         end
         %% version
         meth showVersion
            {self showInfo(AppMan,getLocalEnv('_MOSH_VERSION_VS' $))}
         end 
         meth BuiltinVersion(App GetCmdArgs)
            Spec = record
            Args = {GetCmdArgs Spec}
         in
            case Args.1 of nil then
               {self showVersion}
            else
               raise error(ap(usage 'unexpected arguments')) end
            end
         end
         %% alias / unalias
         meth BuiltinAlias(App GetCmdArgs)
            Spec = plain
            Args = {GetCmdArgs Spec}
         in
            case Args of nil then
               {ForAll {Dictionary.entries self.aliases}
                proc {$ Name#(CmdA#Rest)}
                   {App.showInfo Name#"="#CmdA#" "#Rest}
                end}
            [] [_] then
               raise error(ap(usage 'more than one argument expected')) end
            elseof Name|CmdA|Rest then
               if Rest == nil
               then {Dictionary.put self.aliases {VS2A Name} {VS2A CmdA}#nil}
               else {Dictionary.put self.aliases {VS2A Name} {VS2A CmdA}#
                     {VS2S {Enumerate Rest " "}}}
               end
            end
         end
         meth BuiltinUnalias(App GetCmdArgs)
            Spec = plain
            Args = {GetCmdArgs Spec}
         in
            case Args of nil then
               raise error(ap(usage 'argument(s) required')) end
            [] [Name] then 
               {Dictionary.remove self.aliases {VS2A Name}}
            else
               raise error(ap(usage 'too many arguments')) end
            end
         end
         %% source
         meth BuiltinSource(App GetCmdArgs)
            Spec = record
            Args = {GetCmdArgs Spec}
         in
            case Args.1 of nil then
               raise error(ap(usage 'filename required')) end
            [] [Filename] then
               ExitException = self.exitException
            in
               if {File.exists Filename} then
                  {App.showInfo "`"#Filename#"'..."}
                  try {self doFile(Filename)}
                  catch !ExitException then skip end
               else
                  {App.showInfo "`"#Filename#"' not found."}
               end
            else
               raise error(ap(usage 'too many arguments')) end 
            end
         end
         %% apply / start
         meth Apply(App URL ARGS
                    detach: Detach <= false
                    relink: Relink <= false)
            try
               ThreadID RealURL Functor 
               ResolveVS = {App.getLocalEnv 'RESOLVE'}
               Resolver  = {Resolve.make 'mosh' vs(ResolveVS)}
               case {Resolver.localize URL}
               of old(File) then
                  RealURL = File
                  Functor = {Pickle.load File}
               [] new(File) then
                  RealURL = URL
                  try
                     Functor = {Pickle.load File}
                  finally
                     {OS.unlink File}
                  end
               end
               ModMan = if Relink then {New Module.manager init}
                        else App.modMan end
            in
               if Detach then
                  proc {DoExit ExitStatus}
                     {self showUser('['#ThreadID#'] '#RealURL#' done. ')}
                  end
               in
                  {App startApplication(RealURL ?ThreadID
                                        'functor': Functor
                                        args:      ARGS
                                        modMan:    ModMan
                                        detach:    true
                                        info:      RealURL
                                        onExit:    DoExit)}
                  {self showUser('['#ThreadID#'] '#RealURL#' started.')}
               else
                  Timeout = try {String.toInt {VS2S {App.getLocalEnv 'TIMEOUT'}}}
                            catch _ then false end
                  Exit
               in
                  {App startApplication(RealURL ?ThreadID
                                        'functor': Functor
                                        args:      ARGS
                                        modMan:    ModMan
                                        detach:    true
                                        info:      RealURL
                                        exit:      Exit)}
                  if Timeout == false orelse Timeout == 0 then
                     {Wait Exit}
                  else
                     {WaitOr Exit {Alarm Timeout*1000}}
                     if {IsFree Exit} then
                        {App.showError
                         '%****************** timeout ('#Timeout#' s) ******************\n'#
                         '%*** Thread:   ['#ThreadID#'] '#RealURL#'\n'#
                         '%*** Command:  apply '#{Enumerate URL|ARGS ' '}}
                     else
                        {App putLocalEnv('?' Exit)}
                     end
                  end
               end
            catch error(url(localize...)...) then
               {App.showError URL#": not found"}
            [] error(dp(generic 'bytesource:empty'...)...) then
               {App.showError URL#": not an Oz-application"#
                " (empty byte source)"}
            end
         end
         meth BuiltinApply(App GetCmdArgs)
            Spec = plain
            Args = {GetCmdArgs Spec}
         in
            case Args of nil then
               raise error(ap(usage 'URL required')) end
            [] "relink"|URL|Rest then
               {self Apply(App URL Rest relink:true)}
            [] URL|Rest then
               {self Apply(App URL Rest)}
            end
         end
         meth BuiltinStart(App GetCmdArgs)
            Spec = plain
            Args = {GetCmdArgs Spec}
         in
            case Args of nil then
               raise error(ap(usage 'URL required')) end
            [] "relink"|URL|Rest then
               {self Apply(App URL Rest detach:true relink:true)}
            [] URL|Rest then
               {self Apply(App URL Rest detach:true)}
            end
         end
         %% help
         meth showCommands
            {ForAll {Sort {Dictionary.keys self.commands} Value.'=<'}
             proc {$ CmdName}
                AppMan,printInfo(CmdName#" ")
             end}
            AppMan,showInfo('')
         end
         meth BuiltinHelp(App GetCmdArgs)
            Spec = record(commands(single char:&c type:bool default:false))
            Args = {GetCmdArgs Spec}
         in
            if Args.commands then
               {self showCommands}
            else
               case Args.1 of nil then
                  HelpText  =
                  "Type `?' for a list of all available commands. \n"#
                  "Try `<cmd> --help' to get command-specific help. \n"#
                  "  \n"#
                  "builtin commands: \n"#
                  "  version   Prints some version information.\n"#
                  "  help      Prints this text, or a list of commands.\n"#
                  "  alias     Shows all alias entries, or adds a new one.\n"#
                  "  unalias   Removes an alias entry.\n"#
                  "  source    Processes a script file.\n"#
                  "  start     Starts an Oz-application (URL) asynchronously.\n"#
                  "  apply     like start, but also waits for termination.\n"#
                  "            `apply' is the default command, thus may be omitted.\n"#
                  "  exit      Closes the shell (or disconnects from it).\n"#
                  "arguments: \n"#
                  "  are separated by blanks (\"...\" is NOT supported) \n"#
                  "environment access: \n"#
                  "  $VAR   (terminated by a blank), or\n"#
                  "  ${VAR} \n"#
                  "output redirection:\n"#
                  "  >   >>   >\&   >>\& \n"#
                  "backquotes: \n"#
                  "  supported, but <newline> does not work... \n"#
                  "comments: \n"#
                  "  lines beginning with # or % are ignored \n"
               in
                  {App.showInfo HelpText}
               else
                  raise error(ap(usage 'unexpected arguments')) end
               end
            end
         end
         %%
         meth addBuiltin(CmdName Meth)
            if {Dictionary.member self.commands CmdName}==false then
               proc {P App GetCmdArgs}
                  {DoCmd CmdName App proc {$} {self Meth(App GetCmdArgs)} end}
               end
            in {Dictionary.put self.commands CmdName P}
            end
         end
      end
      
      class LocalShell from Shell

         feat
            stdin
            closeInput
            
         meth init(name:       AppName
                   appMan:     AppMan
                   directArgs: DirectArgs <= nil
                   prompt:     Prompt     <= "> "
                   commands:   Commands   <= {NewDictionary}
                   stdin:      StdIn      <= unit
                   rcFiles:    RCFiles    <= [Home#'/.'#AppName#'rc']
                  )

            ExitException in
            
            if StdIn \= unit then
               self.stdin      = StdIn
               self.closeInput = proc {$} skip end
            else
               self.stdin      = {New TextFile init(name:stdin)}
               self.closeInput = proc {$} {self.stdin close} end
            end
              
            Shell, init(name:       AppName
                        appMan:     AppMan
                        directArgs: DirectArgs
                        prompt:     Prompt
                        commands:   Commands
                        rcFiles:    RCFiles)

            ExitException = self.exitException
            try
               {self showInfo("Warning: ^C kills the entire oz process.")}
               {self showInfo("Type exit to leave.")}
               {self loop(self.stdin)}
            catch !ExitException then
               {self exit}
            end            
         end
         meth closeInput
            {self.closeInput}
         end
         meth exit
            {self showInfo("\nBye...")}
            {self closeInput}
         end
      end

      class RemoteShell from Shell Server AppMan
                           
         prop
            locking
            
         meth init(name:       AppName
                   appMan:     App
                   directArgs: DirectArgs <= nil
                   prompt:     Prompt     <= "> "
                   commands:   Commands   <= {NewDictionary}
                   rcFiles:    RCFiles    <= [Home#'/.'#AppName#'rc']
                   ticketFile: TicketFile
                   pidFile:    PIDFile    <= TicketFile#".pid"
                   errorFile:  ErrorFile  <= false)

            ExitException = {NewName} in
            
            AppMan, init(App)
            
            Server, init(ticketFile: TicketFile
                         pidFile:    PIDFile
                         errorFile:  ErrorFile
                         exitExc:    ExitException)
            
            Shell, init(name:       AppName
                        appMan:     App
                        directArgs: DirectArgs
                        prompt:     Prompt
                        commands:   Commands
                        rcFiles:    RCFiles
                        exitExc:    ExitException)
                      
            Server, listen
         end

         meth stdin(VS)
            {self doVS(VS)}
         end

         meth prompt
            PromptFun in
            {Thread.preempt {Thread.this}}
            if @userMessages \= nil then
               {self output(message('\n'))}
               {ForAll {Reverse @userMessages}
                proc {$ Msg} Server, output(message(Msg#'\n')) end}
               userMessages <- nil
            end
            AppMan, getLocalEnv('PROMPT' ?PromptFun)
            Server, prompt({PromptFun})
         end    
         
         meth builtinExit(App GetCmdArgs)
            Spec = record(force(single char:&f type:bool default:false))
            Args = {GetCmdArgs Spec}
         in
            case Args.1 of nil then
               if AppMan,isEmpty($) orelse Args.force==true
               then raise self.exitException end
               else raise disconnect end
               end
            else
               raise error(ap(usage 'unexpected arguments')) end
            end
         end
         meth exit
            Shell, exit
            {self closeInOut}
         end                               
      end
   end
end





