%%%
%%% shell commands
%%%

functor
   
import
   Compiler Open OPIEnv OS

   XTime(mapToVS)
   at 'x-ozlib://mathweb/share/XTime.ozf'

export
   
%%% OS module

   %% files
   tmpnam:          Tmpnam
   unlink:          Unlink
   stat:            Stat
   %% directories
   chDir:           ChDir
   getDir:          GetDir
   getCWD:          GetCWD
   %% sockets
   getServByName:   GetServByName
   getHostByName:   GetHostByName
   %% time
   time:            Time
   gmTime:          GMTime
   localTime:       LocalTime
   %% environment
   getEnv:          GetEnv
   putEnv:          PutEnv
   %% miscellaneous
   uName:           UName
   system:          System %% uses Open.pipe
   
%%% unix-like shell commands

   sleep:           Sleep
   echo:            Echo
   cat:             Cat

%%% Thread-related commands

   ps:              PS
   suspend:         Suspend
   resume:          Resume
   injectException: InjectException
   terminate:       Terminate
   kill:            Kill

   send:            Send
   
%%% evaluate Oz-expressions & execute statements

   evaluate:        Evaluate
   execute:         Execute

%%% test commands
   
   descriptors:     Descriptors
   dummyThread0:    DummyThread0
   dummyThread1:    DummyThread1
   dummyThread2:    DummyThread2

%%% adding to the 'commands'-Dictionary
   
   addOS:           AddCommandsOS
   addUnix:         AddCommandsUnix
   addThread:       AddCommandsThread
   addTest:         AddCommandsTest
   addAll:          AddCommandsAll
   
define
   
   local
      Compile = Compiler.virtualStringToValue
 
      class TextFile from Open.file Open.text end

      V2VS = Value.toVirtualString

      VS2A = VirtualString.toAtom
      VS2S = VirtualString.toString

      fun {Enumerate Xs Sep}
         fun {DoEnumerate Xs}
            case Xs of nil then ''
            [] X|Xr then
               Sep#X#{DoEnumerate Xr}
            end
         end
      in
         case Xs of nil then
            raise enumerate(emptyList) end
         [] X|Xr then
            X#{DoEnumerate Xr}
         end
      end

      %% infinite loop
      proc {Infinite MilliSeconds}
         proc {Loop}
            {Delay MilliSeconds}
            {Loop}
         end
      in
         {Loop}
      end

      CondSelect = Value.condSelect

      Help = help(single char:&h type:bool default:false)
   in
      
%%%
%%% OS module liftings
%%%

      %%
      %% files
      %%
      
      %% OS.tmpnam
      local
         HelpText =
         "usage:   tmpnam \n"#
         "effect:  Prints a freshly created full pathname. \n"#
         "example: setenv TMP_FILE `tmpnam` \n"#
         "See also OS.tmpnam in the MOZART documentation.\n"
         
         Spec = record(Help)
      in
         proc {Tmpnam App GetCmdArgs}  
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  {App.showInfo {OS.tmpnam}}
               else
                  raise error(ap(usage 'unexpected arguments')) end 
               end
            end
         end
      end

      %% OS.unlink
      local
         HelpText =
         "usage:   rm [-f] <file>... \n"#
         "options: -f , --force : ignore all errors \n"#
         "effect:  Removes the given files. \n"#
         "example: rm -f $TMP_FILE \n"#
         "See also OS.unlink in the MOZART documentation.\n"

         Spec = record(Help
                       force(single char:&f type:bool default:false))
         
         proc {TryUnlink File}
            try {OS.unlink File}
            catch _ then {Delay 1} skip end
         end
      in
         proc {Unlink App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  raise error(ap(usage 'filename required')) end
               elseof Files then
                  if Args.force
                  then {ForAll Files TryUnlink}
                  else {ForAll Files OS.unlink}
                  end
               end
            end
         end
      end

      %% OS.stat
      local
         HelpText =
         "usage:   stat <filename> \n"#
         "effect:  Prints some information about the file. \n"#
         "example: stat /tmp \n"#
         "See also OS.stat in the MOZART documentation.\n"

         Spec = record(Help)
      in
         proc {Stat App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  raise error(ap(usage 'filename required')) end
               [] [File] then
                  {App.show {OS.stat File}}
               else
                  raise error(ap(usage 'too many arguments')) end
               end
            end
         end
      end
      
      %%
      %% directories
      %%
      
      %% OS.chDir
      local
         HelpText = 
         "usage:   cd \n"#
         "         cd <directory> \n"#
         "         Note: ~ is not supported yet! \n"#
         "effect:  Changes the working directory. \n"#
         "         If <directory> is omitted, the value of $HOME is used.\n"#
         "example: cd .. \n"#
         "See also OS.chDir in the MOZART documentation.\n"

         Spec = record(Help)
      in
         proc {ChDir App GetCmdArgs}   
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  {OS.chDir {OS.getEnv "HOME"}}
               [] [Dir] then
                  {OS.chDir Dir}
               else
                  raise error(ap(usage 'too many arguments')) end
               end
            end
         end
      end
      
      %% OS.getDir
      local
         HelpText =
         "usage:   ls \n"#
         "         ls <directory> \n"#
         "effect:  Enumerates the entries of <directory>. \n"#
         "         If <directory> is omitted, the current directory is used. \n"#
         "example: ls ${HOME} \n"#
         "See also OS.getDir in the MOZART documentation.\n"

         Spec = record(Help)
      in 
         proc {GetDir App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               proc {ShowEntries Dir}
                  case {OS.getDir Dir} of nil then skip
                  [] First|Rest then
                     {App.printInfo First} 
                     {ForAll Rest proc {$ Entry}
                                     {App.printInfo "  "#Entry}
                                  end}
                     {App.showInfo ""}
                  end
               end
            in
               case Args.1
               of nil   then {ShowEntries {OS.getCWD}}
               [] [Dir] then {ShowEntries Dir}
               else
                  raise error(ap(usage 'too many arguments')) end
               end
            end
         end
      end

      %% OS.getCWD
      local
         HelpText =
         "usage:   pwd \n"#
         "effect:  Prints the current working directory. \n"#
         "See also OS.getCWD in the MOZART documentation.\n"

         Spec = record(Help)
      in 
         proc {GetCWD App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  {App.showInfo {OS.getCWD}}
               else
                  raise error(ap(usage 'unexpected arguments')) end
               end
            end
         end
      end
      
      %%
      %% sockets
      %%

      %% OS.getServByName
      local
         HelpText =
         "usage:   getport <service> \n"#
         "         getport <service> <protocol> \n"#
         "effect:  Prints the port number of the respective service. \n"#
         "         If <protocol> is omitted, 'tcp' is used. \n"#
         "example: getport http \n"#
         "See also OS.getServByName in the MOZART documentation.\n"

         Spec = record(Help)
      in
         proc {GetServByName App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               proc {DoGetServByName Service Protocol}
                  case {OS.getServByName Service Protocol}
                  of false          then {App.showInfo "Service unknown."}
                  elseof PortNumber then {App.showInfo PortNumber}
                  end
               end
            in
               case Args.1
               of [Service]          then {DoGetServByName Service tcp}
               [] [Service Protocol] then {DoGetServByName Service Protocol}
               else
                  {App.showError "usage: getport <service> <protocol>"} 
               end
            end
         end
      end

      %% OS.getHostByName
      local
         HelpText =
         "usage:   hostinfo <host> ... \n"#
         "effect:  Shows information about the specified hostname.\n"#
         "example: hostinfo $HOST \n"#
         "See also OS.getHostByName in the MOZART documentation.\n"

         Spec = record(Help)
      in
         proc {GetHostByName App GetCmdArgs}   
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  {App.showError "usage: hostinfo <host> ..."}
               elseof Hosts then
                  {ForAll Hosts
                   proc {$ Host}
                      HostEnt = {OS.getHostByName Host} in
                      {App.showInfo "name: "#HostEnt.name}
                      case HostEnt.aliases of nil then skip
                      [] First|Rest then
                         {App.printInfo "aliases: "#First}
                         {ForAll Rest proc {$ Alias}
                                         {App.printInfo ", "#Alias}
                                      end}
                         {App.showInfo ""}
                      end
                      {App.printInfo "address-list: "}
                      case HostEnt.addrList of nil then skip
                      [] First|Rest then
                         {App.printInfo First}
                         {ForAll Rest proc {$ AddrS}
                                         {App.printInfo ", "#AddrS}
                                      end}
                      end
                      {App.showInfo ""}
                   end}
               end
            end
         end
      end
      
      %%
      %% time
      %%
      
      %% OS.time
      local
         HelpText = "See OS.time in the MOZART documentation.\n"
         Spec = record(Help)
      in
         proc {Time App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  {App.showInfo {OS.time}}
               else
                  raise error(ap(usage 'unexpected arguments')) end
               end
            end
         end
      end
      
      %% OS.gmTime
      local
         HelpText = "See OS.gmTime in the MOZART documentation.\n"
         Spec = record(Help)
      in
         proc {GMTime App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  {App.show {OS.gmTime}}
               else
                  raise error(ap(usage 'unexpected arguments')) end
               end
            end
         end
      end
      
      %% OS.localTime
      local
         HelpText = "See OS.localTime in the MOZART documentation.\n"
         Spec = record(Help)
      in
         proc {LocalTime App GetCmdArgs}      
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  {App.show {OS.localTime}}
               else
                  raise error(ap(usage 'unexpected arguments')) end
               end
            end
         end
      end
      
      %%
      %% environment variables
      %%

      %% OS.getEnv
      local
         HelpText =
         "usage:    getenv [-lps] <varname> \n"#
         "effect:   Prints the value of the specified variable. \n"#
         "          If it is not set, 'false' is printed. \n"#
         "options:  -l , --local  : Search starts in mosh's local environment. \n"#
         "          -p , --parent : Search starts in mosh's parent environment. \n"#
         "          -s , --system : Search the system environment only. \n"#
         "          The options are exclusive; 'local' is the default. \n"#
         "          Note: in most cases, parent and system environment are the same. \n"#
         "examples: getenv -s USER \n"#
         "          getenv -l _MOSH_VERSION_VS \n"#
         "See also OS.getEnv in the MOZART documentation. \n"
         
         Spec = record(Help
                       mode(single type:atom('local' parent system)
                            default:'local')
                       'local'(char:&l alias:mode#'local')
                       parent(char:&p alias:mode#parent)
                       system(char:&s alias:mode#system))
      in
         proc {GetEnv App GetCmdArgs}   
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  raise error(ap(usage 'variable name expected')) end
               [] [VarName] then
                  Val in
                  case Args.mode
                  of 'local' then {App.getLocalEnv VarName ?Val}
                  [] parent  then {App.getEnv      VarName ?Val}
                  [] system  then {OS.getEnv       VarName ?Val}
                  end
                  if {IsVirtualString Val}
                  then {App.showInfo Val}
                  else {App.show     Val}
                  end
               else
                  raise error(ap(usage 'too many arguments')) end
               end
            end
         end
      end

      %% OS.putEnv
      local
         HelpText =
         "usage:    setenv [-lps] <varname> <value(s)> \n"#
         "effect:   Sets the specified variable to the given value. \n"#
         "options:  -l , --local  : Change mosh's local environment. \n"#
         "          -p , --parent : Change mosh's parent environment. \n"#
         "          -s , --system : Change the system environment. \n"#
         "          The options are exclusive; 'system' is the default. \n"#
         "          Note: in most cases, parent and system environment are the same. \n"#
         "examples: setenv PATH ${PATH}:/home/${USER}/bin \n"#
         "          setenv -l TIMEOUT 60 \n"#
         "See also OS.putEnv in the MOZART documentation. \n"

         Spec = record(Help
                       mode(single type:atom('local' parent system)
                            default:system)
                       'local'(char:&l alias:mode#'local')
                       parent(char:&p alias:mode#parent)
                       system(char:&s alias:mode#system))
         local
            fun {DoEnum Xs}
               case Xs of nil then ''
               [] X|Xr then ' '#X#{DoEnum Xr} end
            end
         in
            fun {Enumerate Xs}
               case Xs of nil then ''
               [] X|Xr then X#{DoEnum Xr} end
            end
         end
      in
         proc {PutEnv App GetCmdArgs}   
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of VarName|Values then
                  Val = {Filter {VirtualString.toString {Enumerate Values}}
                         fun {$ C} {Member C [&" &']}==false end}
               in
                  case Args.mode
                  of 'local' then {App.putLocalEnv VarName Val}
                  [] parent  then {App.putEnv      VarName Val}
                  [] system  then {OS.putEnv       VarName Val}
                  end
               else
                  {App.showError "usage: setenv [-lps] <varname> <value>"}
               end
            end
         end
      end

      %%
      %% miscellaneous
      %%

      %% OS.uName
      local
         HelpText =
         "usage:   uname [-snrvm | -a] \n"#
         "options: -s , --sysname  : Print OS name (SunOS, Linux, ...) \n"#
         "         -n , --nodename : Print hostname (leibniz, goedel, ...) \n"#
         "         -r , --release  : Print OS release (5.5.1, 2.0.36, ...) \n"#
         "         -v , --version  : Print OS version \n"#
         "         -m , --machine  : Print machine type (sun4u, i686, ...) \n"#
         "         -a , --all      : Print all available information \n"#
         "effect:  Shows some platform information. \n"#
         "         The behaviour should be similar to the unix 'uname' command, \n"#
         "         except when the command is invoked without any options. \n"#
         "example: uname \n"#
         "See also OS.uName in the MOZART documentation. \n"
         
         Spec = record(Help
                       sysname( single char:&s type:bool default:false)
                       nodename(single char:&n type:bool default:false)
                       release( single char:&r type:bool default:false)
                       version( single char:&v type:bool default:false)
                       machine( single char:&m type:bool default:false)
                       all(single char:&a type:bool default:false))
      in
         proc {UName App GetCmdArgs}    
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               S N R V M
            in
               if Args.all then S=true N=true R=true V=true M=true
               else
                  S = Args.sysname
                  N = Args.nodename
                  R = Args.release
                  V = Args.version
                  M = Args.machine
               end
               case Args.1 of nil then
                  Utsname={OS.uName}
               in
                  if S orelse N orelse R orelse V orelse M then
                     if S then {App.printInfo Utsname.sysname#' '}  end
                     if N then {App.printInfo Utsname.nodename#' '} end
                     if R then {App.printInfo Utsname.release#' '}  end
                     if V then {App.printInfo Utsname.version#' '}  end
                     if M then {App.printInfo Utsname.machine#' '}  end
                  else    
                     {App.printInfo Utsname.sysname#' '}
                     {App.printInfo Utsname.release#' '}
                     {App.printInfo '('#Utsname.version#') '}
                     {App.printInfo 'on '}
                     {App.printInfo Utsname.nodename#' '}
                     {App.printInfo '('#Utsname.machine#') '}
                  end
                  {App.showInfo ""}
               else
                  raise error(ap(usage 'unexpected arguments')) end
               end
            end
         end
      end

      %% OS.system --> Open.pipe
      local
         HelpText =
         "usage:    system <command line> \n"#
         "effect:   Executes the command line in a fresh OS shell. \n"#
         "          If the command has not terminated after $TIMEOUT seconds, \n"#
         "          an error message is displayed and the control is returned \n"#
         "          to mosh. You can change the TIMEOUT value by \n"#
         "            setenv -l TIMEOUT <seconds> \n"#
         "\n"#
         "!!!!!!!!! NOTE: Currently, $SHELL must be a tcsh. \n"#
         "\n"#
         "examples: system date \n"#
         "          system xterm & \n"

         class TextPipe from Open.text Open.pipe end
      in
	 proc {DoSystem App Line}
	    proc {DoThread}
	       Shell     = {OS.getEnv "SHELL"}
	       ShellArgs = Line#";echo JOB_DONE"
	       P = {New TextPipe init(cmd:Shell args:["-c" ShellArgs])}
               proc {ReadLoop}
                  case {P getS($)}
                  of false      then skip
                  [] "JOB_DONE" then skip
                  elseof Line then
		     {App.showInfo Line}
                     {ReadLoop}
		  end
               end
	    in
	       try {ReadLoop}
               finally {P close} end
	    end
            ThrID Exit Info = {VS2A Line}
            TimeOut TimeOutVS = {App.getLocalEnv 'TIMEOUT'}
	 in
            if TimeOutVS == false then
               raise undefVar('TIMEOUT') end
	    else
	       TimeOut = {StringToInt {VS2S TimeOutVS}}
	    end
	    {App newThread(DoThread ?ThrID info:Info exit:?Exit)}
            {WaitOr Exit {Alarm TimeOut*1000}}
            if {IsDet Exit}==false then
               {App tryTerminate(ThrID)}
               {App.showError "system: timeout ($TIMEOUT="#TimeOut#" sec.)"}
            end
         end
         proc {System App GetCmdArgs}
            Args = {GetCmdArgs plain}
         in
            case Args of nil then
               raise error(ap(usage 'command required')) end
            [] "-h"|_     then {App.showInfo HelpText}
            [] "--help"|_ then {App.showInfo HelpText}
            else
               {DoSystem App {Enumerate Args " "}}
            end
         end
      end

%%%
%%% unix-like shell commands
%%%

      %% sleep
      local
         HelpText =
         "usage:   sleep <seconds> \n"#
         "effect:  Wait for the specified number of seconds. \n"#
         "example: sleep 5 \n"#
         "See also Time.delay in the MOZART documentation. \n"
         
         Spec = record(Help)
      in
         proc {Sleep App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  raise error(ap(usage 'time interval required (in seconds)')) end
               [] [Seconds] then
                  SecondsS = {VirtualString.toString Seconds} in
                  if {String.isInt SecondsS}
                  then {Delay {StringToInt SecondsS}*1000}
                  else raise error(ap(usage 'integer expected')) end
                  end
               else
                  raise error(ap(usage 'too many arguments')) end
               end
            end
         end
      end
      
      %% echo
      local
         HelpText =
         "usage:   echo <word>... \n"#
         "effect:  Shows all specified words, separated by one space at a time. \n"#
         "example: echo You are working on ${HOST}. \n"
      in
         proc {Echo App GetCmdArgs}
            Args = {GetCmdArgs plain}
         in
            case Args of nil then
               {App.showInfo ""}
            [] "-h"|_     then {App.showInfo HelpText}
            [] "--help"|_ then {App.showInfo HelpText}
            else
               {App.showInfo {Enumerate Args " "}}
            end
         end
      end
      
      %% cat
      local
         HelpText =
         "usage:   cat <file> \n"#
         "effect:  Shows the content of the specified file. \n"#
         "         Note: The file must be a textfile. \n"#
         "example: cat /home/${USER}/.tcshrc \n"
         
         Spec = record(Help)
      in
         proc {Cat App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of [File] then
                  F = {New TextFile init(name:File)}
                  proc {Loop}
                     case {F getS($)} of false then skip
                     elseof Line then
                        {App.showInfo Line}
                        {Loop}
                     end
                  end
               in
                  try
                     {Loop}
                  catch error(kernel(type _ _ 'VirtualString'...)...) then
                     {App.showInfo "*** error: "#File#": not a textfile"}
                  [] error(kernel(type _ _ virtualString...)...) then
                     {App.showInfo "*** error: "#File#": not a textfile"} 
                  finally
                     {F close}
                  end
               else
                  {App.showError "usage: cat <file>"}
               end
            end
         end
      end
         
      %% ps ??? (--> thread-related commands, --> Threads!!!)
      local
         HelpText =
         "usage:   ps [-rt] <threadID>... \n"#
         "options: -r , --recursive : descend to subthreads \n"#
         "         -t , --times     : show the times when the threads were created \n"#
         "effect:  Prints information about the specified threads. \n"#
         "         If no thread-IDs are given, all thread are shown. \n"#
         "         In this case, the default for the -r option is false; \n"#
         "         otherwise, the its default value is true. \n"#
         "example: ps -rt \n"#
         "See also Thread.state and Thread.isRunnable in the MOZART documentation. \n"
         
         Spec = record(Help
                       recursive(single char:&r type:bool default:auto)
                       times(single char:&t type:bool default:false))
         
         proc {DoPS App Threads ThreadIDs Recursive Prefix ShowStartTime}
            if {IsFree ThreadIDs} then
               ThreadIDs = {Threads keys($)}
            end
            {ForAll ThreadIDs
             proc {$ ThreadID}
                Info      = {Threads getInfo(ThreadID $ default:'<unknown>')}
                State     = {Threads state(ThreadID $)}
                Suspended = {Threads tryIsSuspended(ThreadID $)}
                InfoVS
             in
                if {IsVirtualString Info}
                then InfoVS = Info
                else InfoVS = {V2VS Info 3 3}
                end
                if ShowStartTime then
                   T       = {XTime.mapToVS {Threads getStart(ThreadID $)}}
                   StartVS = T.mon#" "#T.day#", "#T.hour#":"#T.min#":"#T.sec
                in
                   {App.printInfo StartVS#"  "}
                end
                {App.printInfo Prefix#ThreadID#": "#InfoVS}
                {App.printInfo " ("#State}
                if Suspended == true then
                   {App.printInfo ",suspended"}
                end
                {App.showInfo ")"}
                if Recursive == true then
                   case {CondSelect {Threads get(ThreadID $)} child false}
                   of false then skip
                   elseof ChildApp then
                      {DoPS App ChildApp _ true Prefix#ThreadID#'.' ShowStartTime}
                   end
                end
             end}
         end
      in
         proc {PS App GetCmdArgs}
            Args = {GetCmdArgs Spec}
            R
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  if Args.recursive == auto
                  then R = false
                  else R = Args.recursive
                  end
                  {DoPS App App _ R '' Args.times}
               elseof ThreadIDs then
                  if Args.recursive == auto
                  then R = true
                  else R = Args.recursive
                  end
                  {DoPS App App {Map ThreadIDs VS2A} R '' Args.times}
               end
            end
         end
      end
      
%%%
%%% Thread module liftings
%%%

      %% Thread.suspend
      local
         HelpText =
         "usage:   suspend [-r] <threadID>... \n"#
         "options: -r , --recursive : descend to subthreads \n"#
         "effect:  Suspends the specified threads. \n"#
         "example: Create a dummy-thread with the 'thread' command, then: \n"#
         "           suspend <n> \n"#
         "         where <n> is its threadID. \n"#
         "See also Thread.suspend in the MOZART documentation. \n"
         
         Spec = record(Help
                       recursive(single char:&r type:bool default:false))

         proc {DoSuspend Threads ThreadIDs Recursive}
            if {IsFree ThreadIDs} then
               ThreadIDs = {Threads keys($)}
            end
            {ForAll ThreadIDs
             proc {$ ThreadID} 
                {Threads suspend(ThreadID)}
                if Recursive == true then
                   case {CondSelect {Threads get(ThreadID $)} child false}
                   of false then skip
                   elseof ChildApp then {DoSuspend ChildApp _ true}
                   end
                end
             end}
         end
      in
         proc {Suspend App GetCmdArgs}   
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  raise error(ap(usage 'ThreadID required')) end
               elseof ThreadIDs then
                  {DoSuspend App {Map ThreadIDs VS2A} Args.recursive}
               end
            end
         end
      end
      
      %% Thread.resume
      local
         HelpText =
         "usage:   resume [-r] <threadID>... \n"#
         "options: -r , --recursive : descend to subthreads \n"#
         "example: Create a dummy-thread with the 'thread' command, \n"#
         "         suspend it with the 'suspend <n>', then:\n"#
         "           resume <n> \n"#
         "See also Thread.resume in the MOZART documentation. \n"

         Spec = record(Help
                       recursive(single char:&r type:bool default:false))
         
         proc {DoResume Threads ThreadIDs Recursive}
            if {IsFree ThreadIDs} then
               ThreadIDs = {Threads keys($)}
            end
            {ForAll ThreadIDs
             proc {$ ThreadID} 
                {Threads resume(ThreadID)}
                if Recursive == true then
                   case {CondSelect {Threads get(ThreadID $)} child false}
                   of false then skip
                   elseof ChildApp then {DoResume ChildApp _ true}
                   end
                end
             end}
         end
      in
         proc {Resume App GetCmdArgs}   
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  raise error(ap(usage 'ThreadID required')) end
               elseof ThreadIDs then
                  {DoResume App {Map ThreadIDs VS2A} Args.recursive}
               end
            end
         end
      end
      
      %% Thread.injectException
      local
         HelpText =
         "usage:   inject <threadID> <exception> \n"#
         "         The exception may be an arbitrary Oz-Expression. \n"#
         "         Its representation is allowed to contain blanks. \n"#
         "effect:  Injects the given exception into the specified thread. \n"#
         "example: Create a dummy-thread with the 'thread' command, then: \n"#
         "           inject <n> test(a b:1) \n"#
         "         where <n> is its threadID. \n"#
         "See also Thread.inject in the MOZART documentation. \n"

         Spec = record(Help)
      in
         proc {InjectException App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  {App.showError "usage: inject <threadID> <exception>"}
               [] [_] then
                  {App.showError "usage: inject <threadID> <exception>"} 
               [] ThreadID|Rest then
                  Exc =
                  try
                     {Compile "local skip in "#{Enumerate Rest " "}#" end"}   
                  catch error(compiler(evalExpression...)...) then
                     raise error(ap(usage
                                    '<exception> must be an Oz-expression.'))
                     end
                  end
               in
                  {App injectException({VS2A ThreadID} Exc)}
               end
            end
         end
      end
      
      %% command 'term':  Thread.terminate
      local
         HelpText =
         "usage:   term [options] <threadID>... \n"#
         "options: -r , --recursive : descend to subthreads \n"#
	 "         -f , --force : ignore errors \n"#
	 "              --range [<startPID> <endPID>]"#
         "         -a , --all : terminate all threads \n"#
         "         -e , --exit <exitcode> : specify the exit code (expression) \n"#
         %%"         -i , --info <info> :\n"#
         "         Note: <exitcode> must not contain blanks. \n"#
         "effect:  Terminates the specified thread(s). \n"#
         "         This is done by injecting an exception \n"#
         "           system(kernel(terminate exit: <exitcode>)) \n"#
         "         into the thread. \n"#
         "         <exitcode> defaults to unit. \n"#
         "example: term -a \n"#
         "See also Thread.terminate in the MOZART documentation. \n"

         Spec = record(Help
		       recursive(single char:&r type:bool  default:false)
		       force(single char:&f type:bool      default:false)
		       range(single         type:list(int) default:nil)
		       exit( single char:&e type:string    default:unit)
		       info( single char:&i type:string    default:unit)
		       all(  single char:&a type:bool      default:false))
	 
         proc {DoTerminate Threads ThreadIDs Exit Info Recursive TryFlag}
            if {IsFree ThreadIDs} then
               ThreadIDs = {Threads keys($)}
            end
            {ForAll ThreadIDs
             proc {$ ThreadID}
                if TryFlag
                then {Threads tryTerminate(ThreadID exit:Exit info:Info)}
                else {Threads terminate(ThreadID exit:Exit info:Info)}
                end
                if Recursive == true then
                   case {CondSelect {Threads get(ThreadID $)} child false}
                   of false then skip
                   elseof ChildApp then
                      {DoTerminate ChildApp _ Exit Info true TryFlag}
                   end
                end
             end}
         end
      in
         proc {Terminate App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               Exit =
               if Args.exit == unit then unit
               else
                  try
                     Exit = {Compile Args.exit}
                  catch error(compiler(evalExpression...)...) then
                     raise error(ap(usage
                                    '<exit> must be an Oz-expression.')) end
                  end
               end
               Info = 
               if Args.info == unit then unit
               else
                  try
                     Info = {Compile Args.info}
                  catch error(compiler(evalExpression...)...) then
                     raise error(ap(usage
                                    '<info> must be an Oz-expression.')) end
                  end
               end
	    in
	       if Args.all==true then
		  {DoTerminate App _ Exit Info Args.recursive Args.force}
	       else
		  if {Length Args.range} == 2 then
		     [MinPID MaxPID] = Args.range
		  in
		     {For MinPID MaxPID 1 proc{$ PID}
					     try
						{DoTerminate App
						 [{VS2A PID}]
						 Exit
						 Info
						 Args.recursive
						 Args.force}
					     catch _ then skip end
					  end}
		  else
		     case Args.1 of nil then
			raise error(ap(usage 'ThreadID required')) end
		     elseof ThreadIDs then
			{DoTerminate App {Map ThreadIDs VS2A}
			 Exit Info Args.recursive Args.force}
		     end
		  end
	       end
	    end
	 end
      end
      
      %% kill
      local
         HelpText =
         "usage:   kill [-ader] <threadID>... \n"#
         "options: -e , --exit <exitcode> : specify the exit code (expression) \n"#
         "         -d , --delay <ms> : set the delay between two successive injections\n"#
         "         -r , --retry <n> : how many times the injection is to be repeated\n"#
         "         <exitcode> defaults to unit, \n"#
         "         the default delay is 1 millisecond, \n"#
         "         the default number of retries is 99. \n"#
         "effect:  Injects exceptions into the specified thread(s), until the thread \n"#
         "         terminates. If it doesn't terminate after n retries, kill gives up.\n"#
         "example: Create a semi-resistent loop with the 'thread1' command,\n"#
         "         then kill it by 'kill <n>', where <n> is its threadID. \n"#
         "WARNING: 'kill' is still buggy. Sometimes it causes mosh to crash.\n"#
         "         If a thread cannot be terminated by the 'term' command, \n"#
         "         you might prefer 'suspend'ing it. \n"
         
         Spec = record(Help
                       exit(single char:&e type:string default:unit)
                       delay(single char:&d type:int default:1)
                       retry(single char:&r type:int default:99)
                       all(single char:&a type:bool default:false))
      in
         proc {Kill App GetCmdArgs}  
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               Exit
               if Args.exit == unit
               then Exit = unit
               else
                  try 
                     Exit = {Compile Args.exit}
                  catch error(compiler(evalExpression...)...) then
                     Exit = {VS2A Args.exit}
                  end
               end
               Delay = Args.delay
               Retry = Args.retry
            in
               if Args.all==true then
                  {App killAll(exit:Exit delay:Delay retry:Retry)}
               else
                  case Args.1 of nil then
                     raise error(ap(usage 'ThreadID required')) end
                  elseof ThreadIDs then
                     {ForAll ThreadIDs
                      proc {$ ThreadID_VS}
                         {App kill({VS2A ThreadID_VS} exit:Exit
                                   delay:Delay retry:Retry)}
                      end}
                  end
               end
            end
         end
      end
           
      %% send
      local
         HelpText =
         "usage:   send [-ops] <threadID> <message>... \n"#
         "options: -o , --object : send the message directly to the object \n"#
         "         -s , --server : send <message> to the server \n"#
         "         -p , --port   : send <message> to the server port \n"#
         "         The options are mutually exclusive; the default is `-s'.\n"#
         "effect:  sends the message (an arbitrary Oz expression) to the \n"#
         "         server object associated with the thread. \n"#
         "         If the thread is not a server thread, an exception is raised.\n"#
         "example: Create a server thread, e.g. with \n"#
         "         eval {App newServer(BaseObject noop _ $ info:test detach:true)} \n"#
         "         showing the threadID of the newly created server thread, and then\n"#
         "         send <threadID> test\n"

         UsageErr = error(ap(usage 'usage: send <threadID> <message>')) 
      in
         proc {Send App GetCmdArgs}
            Args = {GetCmdArgs plain}
         in
            case Args of nil then raise UsageErr end 
            [] "-h"|_       then {App.showInfo HelpText}
            [] "--help"|_   then {App.showInfo HelpText}
            [] "-s"|R       then {Send1 App server R}
            [] "--server"|R then {Send1 App server R}
            [] "-o"|R       then {Send1 App object R}
            [] "--object"|R then {Send1 App object R}
            [] "-p"|R       then {Send1 App port   R}
            [] "--port"|R   then {Send1 App port   R}
            else                 {Send1 App server Args}
            end
         end
         proc {Send1 App Mode MsgArgs}
            case MsgArgs
            of nil then raise UsageErr end
            [] [_] then raise UsageErr end
            [] ThrID|R then
               Env = {Adjoin OPIEnv.full env('App':App
                                             'Show':App.show
                                             'Print':App.print)}
               M   = {Compiler.evalExpression {Enumerate R " "} Env _}
            in
               case Mode
               of object then {App sendObject({VS2A ThrID} M)}
               [] server then {App sendServer({VS2A ThrID} M)}
               [] port   then {App sendPort(  {VS2A ThrID} M)}
               end
            end
         end
      end
      
%%%
%%% test commands
%%%
      
      %% infinite loop
      local
         HelpText =
         "usage:   thread \n"#
         "effect:  Creates a dummy thread realizing an infinite loop.\n"#
         "options: -d , --delay <ms> : set the cycle time in milli-seconds. \n"#
         "         The default cycle time is 1000 ms. \n"#
         "example: thread \n"
         
         Spec = record(Help
                       delay(single char:&d type:int default:1000))
      in
         proc {DummyThread0 App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  ThreadID
                  Info = '<infinite loop>'
                  proc {DoThread}
                     {Infinite Args.delay}
                  end
                  proc {DoExit ExitStatus}
                     {App showInfo('['#ThreadID#'] '#Info#' done. ')}
                  end
               in
                  {App newThread(DoThread ?ThreadID
                                 info:    Info
                                 onExit:  DoExit)}
                  {App showInfo('['#ThreadID#'] '#Info#' started.')}
               else
                  raise error(ap(usage 'unexpected arguments')) end
               end
            end
         end
      end
      
      %% semi-resistent loop
      local
         HelpText =
         "usage:   thread1 \n"#
         "effect:  Creates a dummy thread realizing an infinite loop. \n"#
         "         'term' signals are catched, and the loop is restarted after 1 sec.\n"#
         "options: -d , --delay <ms> : set the cycle time in milli-seconds. \n"#
         "         The default cycle time is 1000 ms. \n"#
         "example: thread1 \n"
         
         Spec = record(Help
                       delay(single char:&d type:int default:1000))
      in
         proc {DummyThread1 App GetCmdArgs}      
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  ThreadID
                  Info = '<semi-resistent loop>'
                  proc {DoThread}
                     try {Infinite Args.delay}
                     catch system(kernel(terminate exit:Exit...)...) then
                        {App.printInfo 'term signal catched (exit:'}
                        {App.print Exit}
                        {App.showInfo ')'}
                        {Delay 1000}
                        {DoThread}
                     end
                  end
                  proc {DoExit ExitStatus}
                     {App showInfo('['#ThreadID#'] '#Info#' done. ')}
                  end
               in
                  {App newThread(DoThread ?ThreadID
                                 info:    Info
                                 onExit:  DoExit)}
                  {App showInfo('['#ThreadID#'] '#Info#' started.')}   
               else
                  raise error(ap(usage 'unexpected arguments')) end
               end
            end
         end
      end
      
      %% resistent loop
      local
         HelpText =
         "usage:   thread2 \n"#
         "effect:  Creates a dummy thread realizing an infinite loop. \n"#
         "         All signals are catched, and the loop is restarted immediately. \n"#
         "         Thus this thread cannot be killed. \n"#
         "options: -d , --delay <ms> : set the cycle time in milli-seconds. \n"#
         "         The default cycle time is 1000 ms. \n"#
         "example: thread2 \n"
         
         Spec = record(Help
                       delay(single char:&d type:int default:1000))
      in
         proc {DummyThread2 App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               case Args.1 of nil then
                  ThreadID
                  Info = '<resistent loop>'
                  proc {DoThread}
                     try {Infinite Args.delay}
                     catch _ then {DoThread} end
                  end
                  proc {DoExit ExitStatus}
                     {App showInfo('['#ThreadID#'] '#Info#' done. ')}
                  end
               in
                  {App newThread(DoThread ?ThreadID
                                 info:    Info
                                 onExit:  DoExit)}
                  {App showInfo('['#ThreadID#'] '#Info#' started.')}
               else
                  raise error(ap(usage 'unexpected arguments')) end
               end
            end
         end
      end

      %% how many descriptors are available ?
      local
         HelpText =
         "usage:   desc \n"#
         "effect:  Shows how many (file) descriptors are available. \n"#
         "         This is done by 'trial and error', thus it may cause errors \n"#
         "         in those threads that are trying to require a new descriptor while\n"#
         "         the command is executed. \n"#
         "example: desc \n"#
         "See also the tcsh built-in 'limit' command. \n"
         
         Spec = record(Help)
      in
         proc {Descriptors App GetCmdArgs}
            Args = {GetCmdArgs Spec}
         in
            if Args.help then {App.showInfo HelpText}
            else
               D = {NewDictionary}
               proc {TryLoop N}
                  S = {New Open.socket init} in
                  {Dictionary.put D N S}
                  {TryLoop N+1}
               end
            in
               try {TryLoop 1}
               catch _ then {App showInfo({Length {Dictionary.keys D}}
                                          #" descriptors available.")}
               finally
                  {ForAll {Dictionary.entries D}
                   proc {$ N#S} thread {S close} {Dictionary.remove D N} end end}
                  {Delay 1000}
                  if {Dictionary.keys D} \= nil then
                     {App showError("Error: "#{Length {Dictionary.keys D}}
                                    #" test-sockets haven't been closed yet.")}
                  end
               end
            end
         end
      end
      
%%%
%%% Compile & Execute
%%%

      local
         HelpText =
         "usage:    eval <expression> \n"#
         "effect:   Evaluates the given <expression> and displays the result. \n"#
         "          Before the string is passed to the compiler, it is wrapped by \n"#
         "          'fun {$} ' ... ' end', thus local variables may be used without \n"#
         "          an explicit 'local' ... 'end'. \n"#
         "\n"#
         "          Note: <expression> may contain blanks. \n"#
         "                If used as separators, they will be treated correctly. \n"#
         "                However, consecutive blanks in a string are treated as if \n"#
         "                they were only a single one."#
         "\n"#
         "examples: eval 1+1 \n"#
         "          eval {Cos 0.0} \n"#
         "          eval {Property.get 'platform.os'} \n"
      in
         proc {Evaluate App GetCmdArgs}
            Args = {GetCmdArgs plain}
         in
            case Args of nil then
               raise error(ap(usage 'expression required')) end
            [] "-h"|_     then {App.showInfo HelpText}
            [] "--help"|_ then {App.showInfo HelpText}
            else
               VS     = "fun {$} "#{Enumerate Args " "}#" end"
               Env    = {Adjoin OPIEnv.full env('App':   App
                                                'Show':  App.show
                                                'Print': App.print)}
               Result = {Compiler.evalExpression VS Env _}
            in
               {App.show {Result}}
            end
         end
      end

      local
         HelpText =
         "usage:    exec <statements> \n"#
         "effect:   Executes the given <statements>. \n"#
         "          Before the string is passed to the compiler, it is wrapped by \n"#
         "          'proc {$} ' ... ' end', thus local variables may be used without \n"#
         "          an explicit 'local' ... 'end'. \n"#
         "\n"#
         "          Note: <statements> may contain blanks. \n"#
         "                If used as separators, they will be treated correctly. \n"#
         "                However, consecutive blanks in a string are treated as if \n"#
         "                they were only a single one."#
         "\n"#
         "examples: exec thread {Show start} {Delay 5000} {Show stop} end \n"#
         "          exec {Browse 42} \n"#
         "          exec {Property.put 'messages.gc' true} \n"
      in
         proc {Execute App GetCmdArgs}
            Args = {GetCmdArgs plain}
         in
            case Args of nil then
               raise error(ap(usage 'statement(s) required')) end 
            [] "-h"|_     then {App.showInfo HelpText}
            [] "--help"|_ then {App.showInfo HelpText}
            else
               VS     = "proc {$} "#{Enumerate Args " "}#" end"
               Env    = {Adjoin OPIEnv.full env('App':App
                                                'Show':App.show
                                                'Print':App.print)}
               Result = {Compiler.evalExpression VS Env _}
            in
               {Result}
            end
         end
      end
   end
   local
      proc {DoCmd CmdName App Proc}
         try
            {Proc}
         catch error(ap(usage S)...) then
            {App.showError CmdName#': '#S}
         [] system(os(A F I S)...) then
            {App.showError A#' error: '#F#': ['#I#'] '#S}
         end
      end
      proc {AddCommand Dict CmdName CmdProc}
         proc {P App GetCmdArgs}
            {DoCmd CmdName App proc {$} {CmdProc App GetCmdArgs} end}
         end
      in
         if {Dictionary.member Dict CmdName}==false then
            {Dictionary.put Dict CmdName P}
         end
      end
   in
      proc {AddCommandsOS Dict}
         %% files
         {AddCommand Dict 'tmpnam'        Tmpnam}
         {AddCommand Dict 'unlink'        Unlink}
         {AddCommand Dict 'rm'            Unlink}
         {AddCommand Dict 'stat'          Stat}
         %% directories
         {AddCommand Dict 'chDir'         ChDir}
         {AddCommand Dict 'cd'            ChDir}
         {AddCommand Dict 'getDir'        GetDir} 
         {AddCommand Dict 'dir'           GetDir}
         {AddCommand Dict 'ls'            GetDir}
         {AddCommand Dict 'getCWD'        GetCWD}
         {AddCommand Dict 'pwd'           GetCWD}
         %% sockets
         {AddCommand Dict 'getServByName' GetServByName}
         {AddCommand Dict 'getport'       GetServByName}
         {AddCommand Dict 'getHostByName' GetHostByName}
         {AddCommand Dict 'hostinfo'      GetHostByName}
         %% time
         {AddCommand Dict 'time'          Time}
         {AddCommand Dict 'gmTime'        GMTime}
         {AddCommand Dict 'localTime'     LocalTime}
         %% environment variables
         {AddCommand Dict 'getEnv'        GetEnv}
         {AddCommand Dict 'getenv'        GetEnv}
         {AddCommand Dict 'putEnv'        PutEnv}
         {AddCommand Dict 'setenv'        PutEnv}
         %% miscellaneous
         {AddCommand Dict 'uName'         UName}
         {AddCommand Dict 'uname'         UName}
         {AddCommand Dict 'system'        System}
         {AddCommand Dict 'os'            System}
      end
      proc {AddCommandsUnix Dict}
         {AddCommand Dict 'cat'           Cat}
         {AddCommand Dict 'echo'          Echo}
         {AddCommand Dict 'sleep'         Sleep}
      end
      proc {AddCommandsThread Dict}
         {AddCommand Dict 'ps'            PS}
         {AddCommand Dict 'threads'       PS}
         {AddCommand Dict 'suspend'       Suspend}
         {AddCommand Dict 'resume'        Resume}
         {AddCommand Dict 'inject'        InjectException}
         {AddCommand Dict 'terminate'     Terminate}
         {AddCommand Dict 'term'          Terminate}
         {AddCommand Dict 'kill'          Kill}
         {AddCommand Dict 'send'          Send}
      end
      proc {AddCommandsCompile Dict}
         {AddCommand Dict 'evaluate'      Evaluate}
         {AddCommand Dict 'eval'          Evaluate}
         {AddCommand Dict 'execute'       Execute}
         {AddCommand Dict 'exec'          Execute}
      end
      proc {AddCommandsTest Dict}
         {AddCommand Dict 'thread'  DummyThread0}
         {AddCommand Dict 'thread0' DummyThread0}
         {AddCommand Dict 'thread1' DummyThread1}
         {AddCommand Dict 'thread2' DummyThread2}
         {AddCommand Dict 'desc'    Descriptors}
      end
      proc {AddCommandsAll Dict}
         {AddCommandsOS      Dict}
         {AddCommandsUnix    Dict}
         {AddCommandsThread  Dict}
         {AddCommandsCompile Dict}
         {AddCommandsTest    Dict}
      end   
   end
end










