%%%
%%% text interface client for mosh, mbase, ...
%%%

functor
   
import
   Connection(take offerUnlimited)
   Fault
   Open(text file)
   OS(getEnv)
   Pickle(load)
   System(gcDo)
   
   Ticket(analyze:AnalyzeTicket)
   at 'x-ozlib://mathweb/share/Ticket.ozf'
   Internet(hostname)
   at 'x-ozlib://mathweb/share/Internet.ozf'
   File(saveVS)
   at 'x-ozlib://mathweb/share/File.ozf'
   Process(sigF)
   at 'x-ozlib://mathweb/share/Process.ozf'   

export
   'class': Client
   connect: RunClient
   getRemotePort: GetRemotePort
   
define

   fun {GetRemotePort Contact}
      case {CondSelect Contact method none}
      of firewall1 then
	 RemotePort 
	 T = {Connection.offerUnlimited RemotePort}
      in	 
	 {File.saveVS T#"\n" Contact.file} %% XXX todo: append
	 {Process.sigF Contact.signal [Contact.pid] _}
	 {WaitOr RemotePort {Alarm 5000}}
	 if {IsDet RemotePort} andthen
	    {IsPort RemotePort} then
	    RemotePort
	 else
	    raise error(dp('no remote port')) end
	 end
      [] none then
	 {Connection.take Contact}
      else raise unknown_method end
      end
   end

   class TextFile from Open.text Open.file end
   
   %% Client:
   %%   AppMan:         the XApplication.manager object
   %%   InputFilename:  where the input stream is to be connected to
   %%   TicketFilename: a URL containing an oz-ticket referring to a
   %%                   server object that defines the following method:
   %%
   %%               connect(inPort:  ?PIn
   %%                       outPort: POut
   %%                       user:    {OS.getEnv 'USER'}
   %%                       host:    Internet.hostname  % full name
   %%                       %%       {OS.getEnv 'HOST'} % nodename
   %%                       display: {OS.getEnv 'DISPLAY'}
   %%                       exit:    ?Exit)}

   class Client
                 
      attr
         App
         InFile
         TicketFile
         %% step 0
         Ticket    : ''
         THost     : ''
         TPort     : ''
         %% step 1
         ServerObj : ''
         %% step 2
         POut      : ''
         
      meth init(appMan:     AppMan
                inputFile:  InputFilename <= stdin
                ticketFile: TicketFilename
               ) 
         App        <- AppMan
         InFile     <- InputFilename
         TicketFile <- TicketFilename
         {self Step0}
      end
      
      %%
      %% step 0: Load ticket from file
      %%
      meth LoadTicket
         {@App.printDebug "Loading file `"#@TicketFile#"'... "}
         Ticket <- {Pickle.load @TicketFile}
         {@App.showDebug "done."}
	 if {IsAtom @Ticket} then
	    {@App.printDebug "Analyzing ticket... "}
	    local
	       T = {AnalyzeTicket @Ticket} in
	       THost <- T.host
	       TPort <- T.port
	    end
	    {@App.showDebug "done."}
	 else
	    THost <- 'localhost'
	    TPort <- '???'
	 end
      end
      %%
      meth Step0
         try
            {self LoadTicket}
            {self Step1}
         catch Err then
            case Err of system(kernel(terminate...)...) then raise Err end
            [] error(url(load URL)...) then
               {@App.showError URL#": file not found"}
            [] error(dp(generic _ S ...)...) then
               {@App.showError @TicketFile#": "#S}
            [] ticket(...) then
               {@App.showError @TicketFile#": oz-ticket expected"}
            else
               {@App.showError "*** ERROR [ticket]:"}
               {@App printExc(Err)}
            end
         end
      end
      
      %%
      %% step 1: Connect to remote server
      %%
      meth Connect(?ServerObj)
         {@App.printDebug "Connecting to "#@THost#" at port "#@TPort#"... "}
         ServerPort = {GetRemotePort @Ticket} in
         ServerObj  = proc {$ M} {Port.send ServerPort M} end
         {@App.showDebug "done."}
         {@App.printDebug "server object: "} {@App.showD ServerObj}
      end
      %%
      meth Step1
         try
            ServerObj <- {self Connect($)}
            {self Step2}
         catch Err then
            case Err of system(kernel(terminate...)...) then raise Err end
            [] error(connection(ticketToDeadSite...)...) then
               {@App.showError @TicketFile#": ticket refers to dead site"}
            [] error(connection(ticketTakeTimeOut...)...) then
               {@App.showError @TicketFile#": timeout while trying to connect"}
            else
               {@App.showError "*** ERROR [remote server]:"}
               {@App printExc(Err)}
            end
         finally
            {@App.printDebug "Garbage collection (-> remote object)... "}
            {System.gcDo}
            {@App.showDebug "done."}
         end
      end
      
      %%
      %% step 2: Initialize output & error streams
      %%
      meth InitOutput(?POut ?OutID ?OutExit) 
         {@App.printDebug "Initializing output & error stream... "}
         local
            SOut
            !POut = {Port.new SOut}
            proc {DoThread}
               {ForAll SOut
                proc {$ Msg}
                   case Msg of false    then {Thread.terminate {Thread.this}}
                   [] ping(X)           then X = unit
                   [] connect(VS...)    then {@App.printError VS}
                   [] disconnect(VS...) then {@App.printError VS}
                   [] error(VS...)      then {@App.printError VS}
                   [] warning(VS...)    then {@App.printError VS}
                   [] message(VS...)    then {@App.printInfo VS}
                   [] prompt(VS...)     then {@App.printInfo VS}
                   [] stdout(VS)        then {@App.printInfo VS}
                   [] stderr(VS)        then {@App.printError VS}
                   end
                end}
            end
            proc {DoError Err}
               {@App.showError "*** OUTPUT ERROR:"}
               {@App printExc(Err)}
            end
            proc {DoExit _}
               {@App.showDebug "`output' done."}                  
            end
         in
            {@App newThread(DoThread ?OutID
                            exit:    ?OutExit
                            info:    'output'
                            onError: DoError
                            onExit:  DoExit)}
         end
         {@App.showDebug "done."}
      end
      %%
      meth Step2
         OutID OutExit in
         try
            POut <- {self InitOutput($ ?OutID ?OutExit)}
            {self Step3}
         catch Err then
            case Err of system(kernel(terminate...)...) then raise Err end
            else
               {@App.showError "*** ERROR [init output]:"}
               {@App printExc(Err)}
            end
         finally
            {@App.printDebug "Shutting down output stream... "}
            {@App tryTerminate(OutID)}
            {Wait OutExit}
            {@App.showDebug "done."}
         end
      end
      
      %%
      %% step 3: Log-on and start input stream readloop
      %%
      meth LogOn(POut ?PIn exit:?Exit)
         {@App.showDebug "Connecting I/O... "}
         {@ServerObj connect(inPort:  ?PIn
                             outPort: POut
                             user:    {OS.getEnv 'USER'}
                             host:    Internet.hostname  % full name
                             %%       {OS.getEnv 'HOST'} % nodename
                             display: {OS.getEnv 'DISPLAY'}
                             exit:    ?Exit)}
         {Wait PIn}
      end
      %%
      meth InitInput(PIn ?InID ?InExit exit:Exit)
         {@App.printDebug "Opening `"#@InFile#"' for reading... "}
         StdIn = {New TextFile init(name:@InFile)}
         {@App.showDebug "done."}
         proc {Watch _ _}
            {@App.showError "*** Server crash. Exiting..."}
            try Exit = error(dp)
            catch _ then skip end
         end
         proc {DoThread}
            case {StdIn getS($)} of false then
               {Thread.terminate {Thread.this}}
            elseof Line then
               try
                  {Port.send PIn Line}
               catch system(dp(...)...) then
                  {Watch _ _}
                  {Thread.terminate {Thread.this}}
               end
            end
            {DoThread}
         end
         proc {DoError Err}
            {@App.showError "*** INPUT ERROR:"} 
            {@App printExc(Err)}
            try Exit = error(input)
            catch _ then skip end 
         end
         proc {DoExit _}
            {Fault.deInstallWatcher PIn Watch _}
            try {Port.send PIn false}
            catch _ then skip end
            {@App.printDebug "closing `"#@InFile#"'... "}
            {StdIn close}
            {@App.showDebug "`input' done."}
         end
      in
         {Fault.installWatcher PIn [permFail] Watch _}
         {@App.printDebug "Initializing input stream... "}
         {@App newThread(DoThread ?InID
                         exit:    ?InExit
                         info:    'input'
                         onError: DoError
                         onExit:  DoExit)}
         {@App.showDebug "done."}
      end
      %%
      meth Step3
         Exit PIn InID InExit in
         try
            {self LogOn(@POut ?PIn exit:?Exit)}
            if PIn == false then
               InID   = false
               InExit = unit
            else
               {self InitInput(PIn ?InID ?InExit exit:Exit)} 
            end
            {@App.showDebug "Waiting for `Exit' signal..."}
            {Wait Exit}
            {@App.showDebug "`Exit' signal detected; connection done."}
         catch Err then
            case Err of system(kernel(terminate...)...) then raise Err end
            else
               {@App.showError "*** ERROR [log-on or init-input]:"}
               {@App printExc(Err)}
            end
         finally
            {@App.showDebug "Shutting down input stream..."}
            {@App tryTerminate(InID)}
            {Wait InExit}
            {@App.showDebug "Shutting down input stream... done."}
            {Delay 1000}
         end
      end

   end % of class Client

   proc {RunClient App TicketFile InFile}
      _ = {New Client init(appMan:App inputFile:InFile ticketFile:TicketFile)}
   end
   
end
