%%%
%%% text server
%%%

functor
   
import
   Application(exit)
   Connection(gate take)
   DPInit %%(getSettings)       [for Mozart 1.2.0]
   DPMisc %%(initIPConnection)  [for Mozart 1.1.0]
   Fault(installWatcher deInstallWatcher)
   OS(getPID system tmpnam uName unlink signal)
   Pickle(save)
   Property(get put)
   System(gcDo)
   Type(ask)
   
   XApplication(wrapper:AppMan)
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   XTime(localTimeStamp)
   at 'x-ozlib://mathweb/share/XTime.ozf'
   File(saveVS loadS)
   at 'x-ozlib://mathweb/share/File.ozf'

export
   'class': Server
   
define

   local 
      V2VS = Value.toVirtualString
      VS2A = VirtualString.toAtom
      VS2S = VirtualString.toString
      
      class TmpFiles
         
         attr Fs: nil
            
         meth registerTmpFile(F)
            Fs<-F|@Fs
         end
         meth removeTmpFiles
            {ForAll @Fs
             proc {$ F} TmpFiles, RemoveTmpFile(F) end}
            Fs<-nil
         end
         meth RemoveTmpFile(F)
            try {OS.unlink F}
            catch _ then skip end
         end
      end
   in
      class Server from AppMan TmpFiles

         prop
            locking
            
         feat
            closeInOut
            port
            ExitExc
            
         attr
            stream:         nil
            queue:          nil
            connection:     false
            watch:          false
            listenerThread: unit
            
         meth init(%% ? name:       AppName
                   %% ? appMan:     App
                   ticketFile: TicketFile
                   pidFile:    PIDFile    <= TicketFile#".pid"
                   errorFile:  ErrorFile  <= false
                   exitExc:    ExitException)

            self.ExitExc  = ExitException
            
            {self initOutputRedirection}
            
            self.closeInOut = proc {$} {self disconnect} end

            local S in
               self.port = {Port.new S}
               stream   <- S
            end

            {self saveTicket(TicketFile)}
            {self savePID(PIDFile)}
            
	    {self initToplevelExceptionHandler(errorFile:ErrorFile)}
         end
         
         meth closeInOut
            {self.closeInOut}
         end

         %% %%%%%%%%%%%%%%%%%% %%
         %% Output Redirection %%
         %% %%%%%%%%%%%%%%%%%% %%

         meth initOutputRedirection
            SharedLock = {NewLock}
            proc {PrintInfoProc VS}
               {Type.ask.virtualString VS}
               {self output(stdout({ByteString.make VS}))}
            end
            proc {PrintErrorProc VS}
               {Type.ask.virtualString VS}
               {self output(stderr({ByteString.make VS}))}
            end
         in
            AppMan, pushOutput(printInfo:  PrintInfoProc
                               printError: PrintErrorProc
                               lockInfo:   SharedLock
                               lockError:  SharedLock)
         end
         
         meth output(X)
            if @connection \= false
            then
               try {Port.send @connection.outPort X}
               catch _ then {self queue(X)} end
            else
               {self queue(X)}
            end
         end
         
         meth queue(Item)
            queue <- Item|@queue
         end
         
         meth rewriteQueue
            ReverseQueue = {Reverse @queue}
         in
            queue <- nil
            {ForAll ReverseQueue
             proc {$ Item}
                case Item
                of stdout(VS) then {self printInfo(VS)}
                [] stderr(VS) then {self printError(VS)}
                [] Other      then {self output(Other)}
                end
             end}
         end       

         %% %%%%%%%%%%%%%%%%%%%%%%%%%% %%
         %% Toplevel Exception Handler %%
         %% %%%%%%%%%%%%%%%%%%%%%%%%%% %%
         
         meth initToplevelExceptionHandler(errorFile:ErrorFile <= false)
            %% Redirect the error messages & make the error handler non-halting
            proc {ToplevelExc Exc}
               `%5`="%%%%%" `%10`=`%5`#`%5`
               Intro = '\n'
               #`%10`#`%10`#`%10`#`%10`#`%10`#`%10`#`%5`#'\n'
	       #"%%% Error on toplevel "#"("#{XTime.localTimeStamp}#")\n"
               #`%10`#`%10`#`%10`#`%10`#`%10`#`%10`#`%5`#'\n'
            in
               lock AppMan,lockError($) then
                  {self printError(Intro)}
                  {self printExc(Exc)}
               end
            end
            proc {ErrHandler Exc}
               case Exc of system(kernel(terminate...)...) then skip
               else
                  if ErrorFile \= false then
                     try AppMan,redirectOutputToFile(
                                   proc {$} {ToplevelExc Exc} end
                                   stderr:append(ErrorFile))
                        %%bug: concurrent output might be redirected, too
                     catch _ then skip end
                  end
                  try {ToplevelExc Exc}
                  catch _ then skip end
               end
            end
         in
            {Property.put 'errors.handler' ErrHandler}
         end

         %% %%%%%%%%%%%%%%%%%% %%
         %% Ticket & PID Files %%
         %% %%%%%%%%%%%%%%%%%% %%
         
	 meth saveTicket(Filename)
	    FW = if {Property.get 'oz.version'} < '1.2.0' 
                 then {DPMisc.initIPConnection info}.firewall
                 else {DPInit.init info _} {DPInit.getSettings}.firewall
                 end
	 in
	    if FW then
	       Tmp = {OS.tmpnam}
	       %% create empty Tmp:
	       %% {File.saveVS '' Tmp}
	       %% {OS.system 'chmod 600 '#Tmp _}
	       proc {SIGUSR1_Handler _}
		  try
		     %% first line of Tmp should be an ozticket now
		     %% read it and remove it from the file
		     %% XXX todo, just read the complete file for now
		     %% XXX and delete it afterwards
		     Line = {Filter {File.loadS Tmp}
			     fun {$ C} {Char.isSpace C}==false end}
		     T = {VS2A Line}
		     {OS.unlink Tmp}
		     %% connect, verify that the remote value is a
		     %% free variable, then bind it to self.port
		     %% XXX todo: review & and improve
		     RemoteVal = {Connection.take T}
		     RemoteVal = self.port
		  in skip 
		  catch E then
		     {self showInfo('\n*** Error in SIGUSR1_Handler:')}
		     {self printExc(E)}
		  end
	       end
	       Contact = contact(method: firewall1
				 file:   Tmp
				 pid:    {OS.getPID}
				 signal: 'USR1'
				)
	    in
	       {OS.signal 'SIGUSR1' SIGUSR1_Handler}
	       {Pickle.save Contact Filename}
	       {self registerTmpFile(Tmp)}
	    else
	       G = {New Connection.gate init(self.port)}
	       T = {G getTicket($)}
	    in
	       {Pickle.save T Filename}
	    end
	    {OS.system 'chmod 600 '#Filename _}
	    {self registerTmpFile(Filename)}
	 end
         meth savePID(Filename)
            {File.saveVS {OS.getPID}#"\n" Filename}
            {self registerTmpFile(Filename)}
         end
         
         %% %%%%%%%%%%%%%%%%%%%% %%
         %% Connect & Disconnect %%
         %% %%%%%%%%%%%%%%%%%%%% %%
         
         meth connect(inPort:  ?PIn
                      outPort: POut
                      user:    User_    <= false
                      host:    Host_    <= false
                      display: Display_ <= false
                      exit:    ?Exit)
               
            User    = if User_    \= false then User_    else "" end
            Host    = if Host_    \= false then Host_    else "" end
            Display = if Display_ \= false then Display_ else "" end
         in
            lock
               {Port.send POut connect("*** mosh @ "#{OS.uName}.nodename#
                                       " (PID="#self.pid#","#
                                       " started "#self.starttimeVS#")\n")}
               if @connection \= false then
                  {self RefuseConnection(POut)}
                  {self WarnConnect(user:User host:Host display:Display)}
                  PIn=false Exit=unit
               else
                  SIn in
                  connection <- connection(inPort:  PIn
                                           outPort: POut
                                           user:    User
                                           host:    Host
                                           display: Display)
                  local
                     proc {Watch _ _}
                        lock
                           connection <- false
                           watch      <- false
                        end
                     end
                  in
                     watch <- Watch
                     {Fault.installWatcher POut [permFail] Watch _}
                  end
                  {self rewriteQueue}
                  PIn = {Port.new SIn}
                  AppMan,putEnv('DISPLAY' {VS2A Display})
                  thread
                     {self ReadLoop(SIn exit:?Exit)}
                  end
               end
            end
         end
         meth RefuseConnection(POut)
            Txt=
            "Already connected to\n"#
            "   $USER    = "#@connection.user#"\n"#
            "   $HOST    = "#@connection.host#"\n"#
            "   $DISPLAY = "#@connection.display#"\n"#
            "Please try again later.\n"
         in
            {Port.send POut error(Txt)}
            {Port.send POut disconnect("*** disconnect ***\n")}
            {Port.send POut false}
         end
	 meth WarnConnect(user:User host:Host display:Display)
	    W = {VS2S AppMan,getLocalEnv('WARN_CONNECT' $)} in
	    if {Map W Char.toLower} == "yes" then
               Warning=
	       "\n*** "#{XTime.localTimeStamp}#
               " request for connection received from \n"#
	       "***    $USER    = "#User#"\n"#
               "***    $HOST    = "#Host#"\n"#
	       "***    $DISPLAY = "#Display#"\n"#
               "*** ... connection refused"
            in
               {self output(warning(Warning#'\n'))}
            end
         end
	 meth ReadLoop(SIn exit:?Exit)
	    TerminateException = {NewName}
	    local
	       ReadLoopThread = {Thread.this}
	    in
	       {OS.signal 'SIGTERM'
		proc {$ _}
		   {Thread.injectException ReadLoopThread TerminateException}
		end}
	    end
	    ExitException = self.ExitExc
	 in
	    try
	       try
		  {self prompt}
		  {ForAll SIn
		   proc {$ VS}
		      if VS == false
		      then raise endOfInput end
		      else {self stdin(VS)} {self prompt}
		      end
		   end}
	       catch !TerminateException then
		  try
		     {self stdin("term -a")}
		     {Delay 1000}
		  finally
		     raise ExitException end
		  end
	       end
	    catch Err then
	       case Err
	       of endOfInput then
		  VS = '*** disconnect (end of input) ***' in
                  {self output(disconnect(VS#'\n' endOfInput))}
		  {self disconnect} Exit = unit
	       [] disconnect then
		  VS = '*** disconnect ***' in
                  {self output(disconnect(VS#'\n'))}
		  {self disconnect} Exit = unit
	       [] !ExitException then
		  VS = '*** exit ***' in
		  {self output(disconnect(VS#'\n' exit))}
                  {self disconnect} Exit = exit
                  {Thread.terminate @listenerThread}
                  %%{Thread.terminate {Thread.this}}
	       else
                  {self output(error('*** error ***\n'))}
                  Exit = Err
                  {Thread.injectException @listenerThread Err}
                  %%raise Err end
               end
            end
         end

         %% to be redefined
         meth stdin(VS) = M
            raise methodNotRedefined(M) end
         end
         meth prompt(VS<="> ")
            {self output(prompt(VS))}
         end
         %%
         
         meth disconnect
            lock
               if @connection \= false then
                  try {Port.send @connection.outPort false}
                  catch _ then skip end
                  {Fault.deInstallWatcher @connection.outPort @watch _}
                  connection <- false
                  watch      <- false
               end
               {System.gcDo}
            end
         end
        
         /*
         meth Watcher
            try 
               {Delay 1000}
               case @watch of false then skip
               elseof W then
                  case @connection of false then skip
                  [] connection(outPort:P...) then
                     Ping Bell={Alarm 10000} in
                     thread
                        try {Port.send P ping(Ping)}
                        catch _ then skip end
                     end
                     {WaitOr Ping Bell}
                     if {IsDet Bell} then
                        lock
                           if @watch==W andthen @connection\=false then
                              connection <- false
                           end
                        end
                     end
                  end
               end
            catch _ then skip end
            Server, Watcher
         end
         */
         
         %% %%%%%%%%%%% %%
         %% Accept Loop %%
         %% %%%%%%%%%%% %%

         meth listen(warnConnect:WarnConnect<=true)
            if WarnConnect
            then AppMan, putLocalEnv('WARN_CONNECT' yes)
            else AppMan, putLocalEnv('WARN_CONNECT' no)
            end
            thread
               listenerThread <- {Thread.this}
               Server, OuterAcceptLoop
            end
            watch <- false
            /*
            thread
               Server, Watcher
            end
            */
         end
         meth InnerAcceptLoop
            M|Sr = @stream in
            stream <- Sr
            case M
            of ping(Ping)   then Ping = unit
            [] connect(...) then {self M}
            else raise unexpected(M) end
            end
            {System.gcDo}
            Server, InnerAcceptLoop
         end
         meth OuterAcceptLoop
            ExitException = self.exitException
         in
            try Server, InnerAcceptLoop
            catch Err then
               case Err of system(kernel(terminate...)...) then
                  {self '_TERMINATE'} {self disconnect} Server,ApplicationExit
               [] !ExitException then
                  {self '_EXIT'}      {self disconnect} Server,ApplicationExit
               [] unexpected(M) then
                  {self '_unexpected'(M)}
               else
                  {self '_ERROR'(Err)}
               end
            end
            Server, OuterAcceptLoop
         end
         meth ApplicationExit(ExitI<=0)
            {self removeTmpFiles}
            {Delay 1000}
            {Application.exit ExitI}
         end
         meth '_TERMINATE'
            {self showInfo('*** TERMINATE ***')}
         end
         meth '_EXIT'
            {self showInfo('*** EXIT ***')}
         end
         meth '_unexpected'(M)
            VS = {V2VS M 5 20} in
            lock AppMan,lockError($) then
               {self showError('')}
               {self showError('*** ERROR in listener thread ***')}
               {self showError('***   unexpected message')}
               {self showError('***   expected: connect(...), ping(_)')}
               {self showError('***   got:      '#VS)}
            end
         end
         meth '_ERROR'(Exc)
            lock AppMan,lockError($) then
               {self showError('')}
               {self showError('*** ERROR in listener thread ***')}
               {self printExc(Exc)}
            end
         end
      end
   end
end



