%%%
%%% mozart shell
%%%

functor
   
import
   Connection(take)
   Open(text file)
   OS(uName getCWD tmpnam system unlink stat)
   Pickle(load save)
   Property(put get)
   Remote(manager)
   Resolve(localize)
   System(gcDo)

   
   XApplication(manager:AppMan)
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   Env(home:Home)
   at 'x-ozlib://mathweb/share/Env.ozf'
   File(saveVS)
   at 'x-ozlib://mathweb/share/File.ozf'
   Shell('local')
   at 'x-ozlib://mathweb/mosh/Shell.ozf'
   TextClient(connect getRemotePort)
   at 'x-ozlib://mathweb/mosh/TextClient.ozf'
   
define
   AppName = 'mosh'

   fun {GetRemotePort Ticketfile}
      {TextClient.getRemotePort {Pickle.load Ticketfile}}
   end

   class DebugAppMan from AppMan 
      meth printDebug(VS)
         AppMan,printDebug('__'#VS)
      end
      meth showDebug(VS)
         AppMan,showDebug('__'#VS)
      end
   end
   
   App = {New DebugAppMan init}
   
   %% args
   Spec = record(command(   single char:&e type:string)
		 batchfile( single char:&b type:string)
                 connect(   single char:&c type:bool default:auto)
		 rcfiles(   single type:list(string) default:[""])
		 fast(      single alias:rcfiles#nil)
		 start(     single char:&s alias:noconnect)
                 daemon(    single type:string)
                 ticketfile(single char:&t type:string optional:true)
		 'local'(   single char:&l type:bool default:false)
		 prompt(    single char:&p type:string default:"> ")
		 debug(     single char:&d type:bool default:false)
		 port(      single type:int default:0)
		 firewall(  single type:bool default:false)
		)
   Args
   try
      Args = {App.getCmdArgs Spec}
   catch error(ap(usage S)...) then
      {App.showError AppName#': '#S}
      {App.exit 1}
   end
   %% end of args
   
   Ticketfile = if {HasFeature Args ticketfile} then Args.ticketfile
                else
                   Name = AppName#"@"#{OS.uName}.nodename in
                   if Home\=false
                   then Home#"/."#Name
                   else {OS.getCWD}#"/"#Name
                   end
                end
   
   RCFiles = {Map Args.rcfiles
              fun {$ RCFile}
                 if RCFile=="" then Home#"/."#AppName#"rc"
                 else RCFile end
              end}

   if Args.debug
   then {App debugOn}
   else {App debugOff}
   end
   
   proc {RemoteConnect Ticketfile InFile}
      {TextClient.connect App Ticketfile InFile}
   end
   
   proc {RemoteStart TicketFile InFile Prompt}
      CWD = {OS.getCWD}
      Stat
      functor
	 RemoteShell
      import
	 Error OS(chDir) 
	 XApplication at 'x-ozlib://mathweb/share/XApplication.ozf'
	 Shell        at 'x-ozlib://mathweb/mosh/Shell.ozf'
      define
	 RemoteApp = {New XApplication.manager init}
         {OS.chDir CWD}
         try
            _ = {New Shell.remote init(name:       AppName
                                       appMan:     RemoteApp
                                       directArgs: Args.1
                                       prompt:     Prompt
                                       ticketFile: TicketFile
				       rcFiles:    RCFiles)}
            Stat = unit
         catch E then
            Stat = {Error.messageToVirtualString {Error.exceptionToMessage E}}
         end
      end
      
      RemoteMan =  if Args.firewall then
		      {New Remote.manager init(detach:   true
					       fork:     sh
					       port:     Args.port
					       firewall: Args.firewall)}
		   else
		      {New Remote.manager init(detach:   true
					       fork:     sh
					       port:     Args.port
					      )}
		   end
   in
      {App.showDebug "Starting detached shell..."}
      {RemoteMan apply(url:App.url RemoteShell)}     %% base-url and Functor
      case Stat of unit then
         {App.showDebug "Starting detached shell... done."}
         {RemoteConnect TicketFile InFile}
      else
         {App.showError "*** ERROR [remote shell]:"}
         {App.printError Stat}
         {App.printDebug "Closing remote manager..."}
         {RemoteMan close}
         {App.showDebug " done."}
      end
   end
   
   proc {Local Prompt}
      {App.showDebug "Starting temporary local shell..."}
      _ = {New Shell.'local' init(name:       AppName
                                  appMan:     App
                                  directArgs: Args.1
                                  prompt:     Prompt
                                  rcFiles:    RCFiles)} 
   end

   class TextFile from Open.text Open.file end
      
   proc {LocalScript File}
      StdIn in
      {App.showDebug "Starting temporary local shell..."}
      {App.showDebug "Opening `"#File#"' for reading..."}
      StdIn = {New TextFile init(name:File)}
      {App.showDebug "Opening `"#File#"' for reading... done."}
      _ = {New Shell.'local' init(name:       AppName
                                  appMan:     App
                                  directArgs: Args.1
                                  prompt:     ''
                                  stdin:      StdIn
                                  rcFiles:    RCFiles)}
      {App.showDebug "Closing `"#File#"'..."}
      {StdIn close}
      {App.showDebug "Closing `"#File#"'... done."}
      {Wait {App empty($)}}
   end

   proc {LocalCommand CommandLine}
      Tmp = {OS.tmpnam}
   in
      {File.saveVS CommandLine Tmp}
      try
         {LocalScript Tmp}
      finally
         {OS.unlink Tmp}
      end
   end

   fun {Exists Ticketfile}
      try
         case {Resolve.localize Ticketfile}
         of old(FullPath) then _={OS.stat FullPath} true
         [] new(FullPath) then {OS.unlink FullPath} true
         end
      catch _ then false end
   end
   fun {Alive File}
      try _={GetRemotePort File} true
      catch _ then false
      finally {System.gcDo} end
   end

   if {HasFeature Args command} then
      {LocalCommand Args.command}
   elseif Args.'local' == true then
      {Local Args.prompt}
   elseif {HasFeature Args daemon} then
      if {Alive Ticketfile}
      then {App.showError Ticketfile#": "#AppName#" is alive."}
      else {RemoteStart Ticketfile Args.daemon Args.prompt}
      end
   elseif {HasFeature Args batchfile} then
      {LocalScript Args.batchfile}
   else
      case Args.connect
      of true then
         {RemoteConnect Ticketfile stdin}
      [] auto then
         if {Exists Ticketfile}
         then {RemoteConnect Ticketfile stdin}
         else {RemoteStart Ticketfile stdin Args.prompt}
         end
      [] false then
         try
	    _ = {GetRemotePort Ticketfile}
            {System.gcDo}
            {App.showError Ticketfile#": oz-ticket refers to running process."}
            {App.exit 1}
         catch _ then {Delay 1000} skip end
         {RemoteStart Ticketfile stdin Args.prompt}
      end
   end
   {App.exit 0}
end


